import copy
import string
from typing import List, Optional, Tuple

import nltk
import numpy as np

from . import assertion
from . import constants


# region Text Transformations
class SpaceOutDecorators:
    def __call__(self, x: str) -> str:
        return constants.DECORATOR_GROUP_REGEX.sub(r" \1 ", x)


class DoubleQuotesToSingle:
    def __call__(self, x: str) -> str:
        return constants.DOUBLE_QUOTE_REGEX.sub("'", x)


class UnquoteText:
    def __call__(self, x: str) -> str:
        return constants.QUOTED_TEXT_REGEX.sub(r" \1 ", x)


class LowerText:
    def __call__(self, x: str) -> str:
        return str(x).lower()


class WordTokenize:
    def __call__(self, x: str) -> List[str]:
        return nltk.tokenize.word_tokenize(x)


class FilterMeaninglessTokens:
    def __call__(self, x: List[str]) -> List[str]:
        return [y for y in x if constants.MEANINGLESS_REGEX.fullmatch(y) is None]


class BasicPradoTransform:
    def __init__(self):
        self._sequence = [
            SpaceOutDecorators(),
            DoubleQuotesToSingle(),
            UnquoteText(),
            LowerText(),
            WordTokenize(),
            FilterMeaninglessTokens(),
        ]

    def __call__(self, x: str) -> List[str]:
        for transform in self._sequence:
            x = transform(x)

        if len(x) == 0:
            x = [""]

        return x


# endregion


# region Stochastic Augmentation
class StochasticAugmentation:
    def _check_init_args(probability: float):
        assertion.assert_type("probability", probability, float)

        assertion.assert_in_range(
            "probability", probability, (0.0, 1.0), inclusivity=(True, True)
        )

        return probability

    def __init__(self, probability: float):
        probability = StochasticAugmentation._check_init_args(probability=probability)

        self._probability = probability

    # region Properties
    @property
    def probability(self) -> float:
        return self._probability

    @property
    def p(self) -> float:
        return self.probability

    # endregion

    def _roll(self) -> bool:
        return np.random.uniform(low=0.0, high=1.0) < self.probability


class StochasticInsertionAugmentation(StochasticAugmentation):
    @staticmethod
    def _check_init_args(alphabet: Optional[str]):
        if alphabet is None:
            alphabet = string.ascii_lowercase

        assertion.assert_type("alphabet", alphabet, str)

        assertion.assert_not_empty("alphabet", alphabet)

        return alphabet

    def __init__(self, probability: float, alphabet: Optional[str] = None):
        super().__init__(probability=probability)

        alphabet = StochasticInsertionAugmentation._check_init_args(alphabet=alphabet)

        self._alphabet = copy.deepcopy(alphabet)

    # region Properties
    @property
    def alphabet(self):
        return self._alphabet

    # endregion

    def __call__(self, x: str) -> str:
        if self._roll():
            char_to_insert = self.alphabet[np.random.randint(len(self.alphabet))]

            if len(x) == 0:
                return char_to_insert

            breaking_point = int(np.random.randint(len(x)))

            return x[:breaking_point] + char_to_insert + x[breaking_point:]

        return x


class StochasticDeletionAugmentation(StochasticAugmentation):
    def __init__(self, probability: float):
        super().__init__(probability=probability)

    def __call__(self, x: str) -> str:
        if len(x) != 0 and self._roll():
            if len(x) == 1:
                return ""
            else:
                index_to_remove = np.random.randint(len(x))

                return x[:index_to_remove] + x[index_to_remove + 1 :]

        return x


class StochasticSwapAugmentation(StochasticAugmentation):
    def __init__(self, probability: float):
        super().__init__(probability=probability)

    def __call__(self, x: str) -> str:
        if len(x) > 1 and self._roll():
            indices_to_swap = np.random.choice(2, 2, replace=False)

            t = list(x)
            t[indices_to_swap[0]], t[indices_to_swap[1]] = (
                t[indices_to_swap[1]],
                t[indices_to_swap[0]],
            )

            return "".join(t)

        return x


class BasicPradoAugmentation:
    @staticmethod
    def _check_init_args(
        insertion_probability: Optional[float],
        deletion_probability: Optional[float],
        swap_probability: Optional[float],
    ):
        if insertion_probability is None:
            insertion_probability = 0.0

        if deletion_probability is None:
            deletion_probability = 0.0

        if swap_probability is None:
            swap_probability = 0.0

        assertion.assert_type("insertion_probability", insertion_probability, float)
        assertion.assert_type("deletion_probability", deletion_probability, float)
        assertion.assert_type("swap_probability", swap_probability, float)

        return insertion_probability, deletion_probability, swap_probability

    def __init__(
        self,
        insertion_probability: Optional[float] = None,
        deletion_probability: Optional[float] = None,
        swap_probability: Optional[float] = None,
    ):
        (
            insertion_probability,
            deletion_probability,
            swap_probability,
        ) = BasicPradoAugmentation._check_init_args(
            insertion_probability=insertion_probability,
            deletion_probability=deletion_probability,
            swap_probability=swap_probability,
        )

        self._sequence = [
            StochasticInsertionAugmentation(probability=insertion_probability),
            StochasticDeletionAugmentation(probability=deletion_probability),
            StochasticSwapAugmentation(probability=swap_probability),
        ]

    def __call__(self, x: str) -> str:
        for transform in self._sequence:
            x = transform(x)

        return x


# endregion



def assert_type(name, value, expected_type):
    try:
        list(expected_type)
        type_name = (
            ", ".join([x.__name__ for x in expected_type[:-1]])
            + " or "
            + expected_type[-1].__name__
        )
    except TypeError:
        type_name = expected_type.__name__

    if not isinstance(value, expected_type):
        raise TypeError(
            f"Expected argument {name} to be of type {type_name}, instead it is "
            f"{type(value).__name__}."
        )


def assert_positive(name, value):
    if value <= 0:
        raise ValueError(f"Expected argument {name} to be positive, but it is {value}.")



import torch


def pad_projections(batch):
    tokens, labels = zip(*batch)

    max_token_length = max(len(token_list) for token_list in tokens)
    tokens = [
        token_list + [""] * (max_token_length - len(token_list))
        for token_list in tokens
    ]

    return tokens, torch.tensor(labels)



import torch
from torch import nn


class ProjectionOperator:
    pass


class PradoProjectionOperator(nn.Module, ProjectionOperator):
    def __init__(self):
        super().__init__()

        self._linear = nn.Linear(in_features=2, out_features=1, bias=False)

        # We set the weights explicitly so we get the mapping:
        #   00 ->  0
        #   01 ->  1
        #   10 -> -1
        #   11 ->  0
        # Also, we automatically do float computation since we'll
        # need floats later, and it's efficient on the GPU.
        with torch.no_grad():
            self._linear.weight = nn.Parameter(
                torch.tensor(
                    [[-1, 1]],
                    dtype=torch.float,
                )
            )

        # The paper doesn't use learnable projection operators
        for param in self.parameters():
            param.requires_grad = False

    def forward(self, x):
        return self._linear(x)



import copy
from typing import Optional

import torch
from torch import nn

from masked_convolution import MaskedConvolution


class SkipgramConvolutionConfig:
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        embedding_length: int,
        skipgram_pattern,
    ):
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.embedding_length = embedding_length
        self.skipgram_pattern = skipgram_pattern


class SkipgramConvolution(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        embedding_length: int,
        skipgram_pattern: str,
        config: Optional[SkipgramConvolutionConfig] = None,
    ):
        super().__init__()

        if config is None:
            config = SkipgramConvolutionConfig(
                in_channels=in_channels,
                out_channels=out_channels,
                embedding_length=embedding_length,
                skipgram_pattern=skipgram_pattern,
            )

        self._config = copy.deepcopy(config)

        with torch.no_grad():
            self._convolution_mask = torch.ones(
                (len(self.skipgram_pattern), self.d), dtype=torch.float32
            )

            for i, char in enumerate(skipgram_pattern):
                if char == "0":
                    self._convolution_mask[i] *= 0

        self._conv = MaskedConvolution(
            conv_layer=nn.Conv2d(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=self._convolution_mask.shape,
                stride=1,
                padding=(self._convolution_mask.shape[0] // 2, 0),
                bias=True,
                padding_mode="zeros",
            ),
            mask=self._convolution_mask,
        )

    # region Properties
    @property
    def in_channels(self):
        return self._config.in_channels

    @property
    def out_channels(self):
        return self._config.out_channels

    @property
    def embedding_length(self):
        return self._config.embedding_length

    @property
    def d(self):
        return self.embedding_length

    @property
    def skipgram_pattern(self):
        return self._config.skipgram_pattern

    # endregion

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # (B, in_features, N, d) -> (B, out_features, N, d)
        return self._conv(x)



import copy
from typing import Any, Dict, List, Optional

import bitarray
from datasketch import MinHash
import farmhash
import numpy as np
import torch

from . import assertion
from .projection_operator import PradoProjectionOperator


class Projector:
    def project(self, x: str):
        raise NotImplementedError

    def __call__(self, x):
        raise NotImplementedError


class PradoProjectorConfig:
    __feature_length_key = "feature_length"

    @staticmethod
    def _check_init_args(feature_length: int):
        assertion.assert_type("feature_length", feature_length, int)

        assertion.assert_positive("feature_length", feature_length)

        return feature_length

    def __init__(self, feature_length: int):
        feature_length = PradoProjectorConfig._check_init_args(
            feature_length=feature_length
        )

        self.feature_length = feature_length

    def serialize(self):
        return {PradoProjectorConfig.__feature_length_key: self.feature_length}

    @staticmethod
    def from_dict(dictionary: Dict[str, Any]) -> "PradoProjectorConfig":
        feature_length = dictionary.get(PradoProjectorConfig.__feature_length_key)

        return PradoProjectorConfig(feature_length=feature_length)


class PradoProjector(Projector):
    def __init__(
        self,
        feature_length: int = None,
        config: Optional[PradoProjectorConfig] = None,
    ):
        super().__init__()

        if config is None:
            config = PradoProjectorConfig(feature_length=feature_length)

        self._config = copy.deepcopy(config)
        self._hashobj = MinHash(num_perm=self.n_permutations, hashfunc=farmhash.hash32)
        self._projection_operator = PradoProjectionOperator()

        self._vectorized_projection = np.vectorize(self.project, signature="()->(n)")

    # region Properties
    @property
    def feature_length(self) -> int:
        return self._config.feature_length

    @property
    def B(self) -> int:
        return self.feature_length

    @property
    def n_permutations(self) -> int:
        return (2 * self.B + 32 - 1) // 32

    # endregion

    def project(self, x: str):
        self._hashobj.clear()
        self._hashobj.update(x)

        # (4 * n_permutations, )
        token_as_bytes = b"".join(
            int(x).to_bytes(4, "big") for x in self._hashobj.digest()
        )

        # (32 * n_permutations, )
        token_as_bits = bitarray.bitarray()
        token_as_bits.frombytes(token_as_bytes)

        # (2B, ) - MinHash can give us larger hashes than
        # we need. It is recommended to set B up so this
        # doesn't destroy/skip data. In other words, B should
        # be a multiplier of 16.
        return torch.tensor(token_as_bits[: 2 * self.B], dtype=torch.float)

    def __call__(self, x: List) -> torch.Tensor:
        # Can be anything, (Any, N[str]) -> (Any, N, 2B)
        token_features = self._vectorized_projection(x)
        token_features = torch.tensor(token_features, dtype=torch.float)

        # (Any, N, 2B) -> (Any, N, B, 2)
        token_features = torch.reshape(
            token_features, (*token_features.shape[:-1], -1, 2)
        )

        # (Any, N, B, 2) -> (Any, N, B, 1)
        fingerprint = self._projection_operator(token_features)

        # (Any, N, B, 1) -> (Any, N, B)
        fingerprint = torch.squeeze(fingerprint, dim=-1)

        return fingerprint



def assert_type(name, value, expected_type):
    try:
        list(expected_type)
        type_name = (
            ", ".join([x.__name__ for x in expected_type[:-1]])
            + " or "
            + expected_type[-1].__name__
        )
    except TypeError:
        type_name = expected_type.__name__

    if not isinstance(value, expected_type):
        raise TypeError(
            f"Expected argument {name} to be of type {type_name}, instead it is "
            f"{type(value).__name__}."
        )


def assert_callable(name, value):
    if not callable(value):
        raise TypeError(
            f"Expected argument {name} to be callable, instead it is "
            f"{type(value).__name__}"
        )


def assert_iterable(name, value):
    try:
        iter(value)
    except TypeError:
        raise TypeError(
            f"Expected argument {name} to be iterable, instead it is "
            f"{type(value).__name__}."
        )


def assert_positive(name, value):
    if value <= 0:
        raise ValueError(f"Expected argument {name} to be positive, but it is {value}.")


def assert_non_negative(name, value):
    if value < 0:
        raise ValueError(
            f"Expected argument {name} to be non-negative, but it is {value}."
        )


def assert_all_int(name, value):
    for i, element in enumerate(value):
        try:
            assert_type(name, value, int)
        except ValueError:
            raise ValueError(
                f"Expected argument {name} to only contain ints, but found a "
                f"{type(element).__name__} at index {i}."
            )


def assert_all_callable(name, value):
    for i, element in enumerate(value):
        try:
            assert_all_callable(name, value)
        except ValueError:
            raise ValueError(
                f"Expected argument {name} to only contain callables, but found a "
                f"{type(element).__name__} at index {i}."
            )


def assert_all_non_negative(name, value):
    for i, element in enumerate(value):
        try:
            assert_non_negative(name, value)
        except ValueError:
            raise ValueError(
                f"Expected argument {name} to only contain non-negative elements, but "
                f"found {element} at index {i}."
            )



from typing import Tuple, Union


def assert_type(name: str, value, expected_type):
    try:
        list(expected_type)
        type_name = (
            ", ".join([x.__name__ for x in expected_type[:-1]])
            + " or "
            + expected_type[-1].__name__
        )
    except TypeError:
        type_name = expected_type.__name__

    if not isinstance(value, expected_type):
        raise TypeError(
            f"Expected argument {name} to be of type {type_name}, instead it is "
            f"{type(value).__name__}."
        )


def assert_callable(name, value):
    if not callable(value):
        raise TypeError(
            f"Expected argument {name} to be callable, instead it is "
            f"{type(value).__name__}"
        )


def assert_not_empty(name: str, value):
    if len(value) == 0:
        raise ValueError(f"Expected argument {name} to not be empty.")


def assert_non_negative(name, value):
    if value < 0:
        raise ValueError(
            f"Expected argument {name} to be non-negative, but it is {value}."
        )


def assert_in_range(
    name: str,
    value,
    value_range: Tuple[Union[float, int], Union[float, int]],
    inclusivity: Tuple[bool, bool] = (True, False),
):
    s = ["or equal to" if inclusivity[i] else "than" for i in range(2)]

    if value < value_range[0] or (value == value_range[0] and not inclusivity[0]):
        raise ValueError(
            f"Expected argument {name} to be less {s[0]} {value_range[0]}, instead it "
            f"is {value}."
        )
    elif value > value_range[1] or (value == value_range[1] and not inclusivity[1]):
        raise ValueError(
            f"Expected argument {name} to be greater {s[1]} {value_range[1]}, "
            f"instead it is {value}."
        )


def assert_all_int(name, value):
    for i, element in enumerate(value):
        try:
            assert_type(name, element, int)
        except ValueError:
            raise ValueError(
                f"Expected argument {name} to only contain ints, but found a "
                f"{type(element).__name__} at index {i}."
            )


def assert_all_float(name, value):
    for i, element in enumerate(value):
        try:
            assert_type(name, element, float)
        except ValueError:
            raise ValueError(
                f"Expected argument {name} to only contain floats, but found a "
                f"{type(element).__name__} at index {i}."
            )


def assert_all_callable(name, value):
    for i, element in enumerate(value):
        try:
            assert_callable(name, element)
        except ValueError:
            raise ValueError(
                f"Expected argument {name} to only contain callables, but found a "
                f"{type(element).__name__} at index {i}."
            )


def assert_all_non_negative(name, value):
    for i, element in enumerate(value):
        try:
            assert_non_negative(name, element)
        except ValueError:
            raise ValueError(
                f"Expected argument {name} to only contain non-negative elements, but "
                f"found {element} at index {i}."
            )



import copy
from typing import Any, Dict, Optional

import torch
from torch import nn

from . import assertion


class PradoEmbeddingConfig:
    __in_features_key = "in_features"
    __out_features_key = "out_features"
    __bias_key = "bias"

    @staticmethod
    def _check_init_args(in_features: int, out_features: int, bias: bool):
        assertion.assert_type("in_features", in_features, int)
        assertion.assert_type("out_features", out_features, int)
        assertion.assert_type("bias", bias, bool)

        assertion.assert_positive("in_features", in_features)
        assertion.assert_positive("out_features", out_features)

        return in_features, out_features, bias

    def __init__(self, in_features: int, out_features: int, bias: bool):
        in_features, out_features, bias = PradoEmbeddingConfig._check_init_args(
            in_features=in_features, out_features=out_features, bias=bias
        )

        self.in_features = in_features
        self.out_features = out_features
        self.bias = bias

    def serialize(self):
        return {
            PradoEmbeddingConfig.__in_features_key: self.in_features,
            PradoEmbeddingConfig.__out_features_key: self.out_features,
            PradoEmbeddingConfig.__bias_key: self.bias,
        }

    @staticmethod
    def from_dict(dictionary: Dict[str, Any]) -> "EmbeddingConfig":
        in_features = dictionary.get(PradoEmbeddingConfig.__in_features_key)
        out_features = dictionary.get(PradoEmbeddingConfig.__out_features_key)
        bias = dictionary.get(PradoEmbeddingConfig.__bias_key)

        return PradoEmbeddingConfig(
            in_features=in_features, out_features=out_features, bias=bias
        )


class PradoEmbedding(nn.Module):
    def __init__(
        self,
        in_features: int = None,
        out_features: int = None,
        bias: bool = False,
        config: Optional[PradoEmbeddingConfig] = None,
    ):
        super().__init__()

        if config is None:
            config = PradoEmbeddingConfig(
                in_features=in_features, out_features=out_features, bias=bias
            )

        self._config = copy.deepcopy(config)

        self._linear = nn.Linear(
            in_features=self.in_features, out_features=self.out_features, bias=self.bias
        )

    # region Properties
    @property
    def in_features(self):
        return self._config.in_features

    @property
    def out_features(self):
        return self._config.out_features

    @property
    def bias(self):
        return self._config.bias

    # endregion

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self._linear(x)



import copy
from typing import Optional

from torch import nn

from .embeddings.embedding import PradoEmbedding
from .projections.projector import PradoProjector


class ProjectedEmbeddingLayerConfig:
    # TODO Serialization
    __feature_length_key = "feature_length"
    __embedding_length_key = "embedding_length"

    def __init__(self, feature_length: int, embedding_length: int):
        # There is no need to check these since they will be
        # automatically checked when subcomponents are created.
        self.feature_length = feature_length
        self.embedding_length = embedding_length


class ProjectedEmbeddingLayer(nn.Module):
    def __init__(
        self,
        feature_length: int = None,
        embedding_length: int = None,
        config: Optional[ProjectedEmbeddingLayerConfig] = None,
    ):
        super().__init__()

        if config is None:
            # The reason we use these indirect values is because
            # they are checked and corrected. It might seem
            # counterintuitive to initialize the config after
            # member variables, but this saves us from
            # unnecessary validation.
            config = ProjectedEmbeddingLayerConfig(
                feature_length=feature_length,
                embedding_length=embedding_length,
            )

        self._config = copy.deepcopy(config)

        self._projection = PradoProjector(feature_length=self.B)
        self._embedding = PradoEmbedding(in_features=self.B, out_features=self.d)

        # Do this to ensure that values which were corrected are
        # correctly corrected in this config as well.
        self._config.feature_length = self._projection.feature_length
        self._config.embedding_length = self._embedding.out_features

    # region Properties
    @property
    def feature_length(self):
        return self._config.feature_length

    @property
    def B(self):
        return self.feature_length

    @property
    def embedding_length(self):
        return self._config.embedding_length

    @property
    def d(self):
        return self.embedding_length

    # endregion

    def forward(self, x):
        projection = self._projection(x)
        projected_embedding = self._embedding(projection)

        return projected_embedding



import copy
from typing import List, Optional

import torch
from torch import nn

from .projected_embedding import ProjectedEmbeddingLayer
from .convolutions.projected_attention import ProjectedAttentionLayer


class PradoCoreConfig:
    def __init__(
        self,
        feature_length: int,
        embedding_length: int,
        dropout: float,
        out_channels: int,
        skipgram_patterns: List[int],
        out_features: int,
    ):
        self.feature_length = feature_length
        self.embedding_length = embedding_length
        self.dropout = dropout
        self.out_channels = out_channels
        self.skipgram_patterns = copy.deepcopy(skipgram_patterns)
        self.out_features = out_features


class PradoCore(nn.Module):
    def __init__(
        self,
        feature_length: int,
        embedding_length: int,
        dropout: float,
        out_channels: int,
        skipgram_patterns: List[str],
        out_features: int,
        config: Optional[PradoCoreConfig] = None,
    ):
        super().__init__()

        if config is None:
            config = PradoCoreConfig(
                feature_length=feature_length,
                embedding_length=embedding_length,
                dropout=dropout,
                out_channels=out_channels,
                skipgram_patterns=skipgram_patterns,
                out_features=out_features,
            )

        self._config = copy.deepcopy(config)

        self._projected_embedding_layer = ProjectedEmbeddingLayer(
            feature_length=feature_length, embedding_length=embedding_length
        )

        self._projected_attention_layers = nn.ModuleList(
            [
                ProjectedAttentionLayer(
                    in_channels=1,
                    out_channels=self.out_channels,
                    embedding_length=self.embedding_length,
                    skipgram_pattern=skipgram_pattern,
                )
                for skipgram_pattern in self.skipgram_patterns
            ]
        )

        self._decoder = nn.Linear(
            in_features=self.encoder_out_features,
            out_features=self.out_features,
            bias=True,
        )

        self._dropout = nn.Dropout(p=self.dropout)
        self._tanh = nn.Tanh()

    # region Properties
    @property
    def feature_length(self):
        return self._config.feature_length

    @property
    def B(self):
        return self.feature_length

    @property
    def embedding_length(self):
        return self._config.embedding_length

    @property
    def d(self):
        return self.embedding_length

    @property
    def dropout(self):
        return self._config.dropout

    @property
    def out_channels(self):
        return self._config.out_channels

    @property
    def skipgram_patterns(self):
        return self._config.skipgram_patterns

    @property
    def encoder_out_features(self):
        return sum([x.out_channels for x in self._projected_attention_layers])

    @property
    def out_features(self):
        return self._config.out_features

    # endregion

    def forward(self, x: torch.Tensor):
        # (batch_size, N, 2B) -> (batch_size, N, d)
        projections = self._projected_embedding_layer(x)
        projections = self._tanh(projections)
        projections = self._dropout(projections)

        # (batch_size, N, d) -> (batch_size, 1, N, d)
        projections = projections.reshape(
            (projections.shape[0], 1, *projections.shape[1:])
        )

        # (n_convolutions, batch_size, out_channels)
        #
        # n_convolutions is len(self._projected_attention_layers)
        features = [layer(projections) for layer in self._projected_attention_layers]

        # (n_convolution, batch_size, out_channels) ->
        # (batch_size, encoder_out_features)
        flattened_features = torch.cat(features, dim=1)
        flattened_features = self._tanh(flattened_features)

        # (batch_size, encoder_out_features) ->
        # (batch_size, out_features)
        decoded = self._decoder(flattened_features)

        return decoded



import copy
from prado.layers.convolutions.skipgram_convolution import SkipgramConvolution
from typing import Optional

import torch
from torch import nn


class ProjectedAttentionLayerConfig:
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        embedding_length: int,
        skipgram_pattern: str,
    ):
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.embedding_length = embedding_length
        self.skipgram_pattern = skipgram_pattern


class ProjectedAttentionLayer(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        embedding_length: int,
        skipgram_pattern: str,
        config: Optional[ProjectedAttentionLayerConfig] = None,
    ):
        super().__init__()

        if config is None:
            config = ProjectedAttentionLayerConfig(
                in_channels=in_channels,
                out_channels=out_channels,
                embedding_length=embedding_length,
                skipgram_pattern=skipgram_pattern,
            )

        self._config = copy.deepcopy(config)

        self._projected_feature_layer = SkipgramConvolution(
            in_channels=self.in_channels,
            out_channels=self.out_channels,
            embedding_length=self.embedding_length,
            skipgram_pattern=self.skipgram_pattern,
        )

        self._attention_layer = SkipgramConvolution(
            in_channels=self.in_channels,
            out_channels=self.out_channels,
            embedding_length=self.embedding_length,
            skipgram_pattern=self.skipgram_pattern,
        )

        self._softmax = nn.Softmax(dim=-1)

    # region Properties
    @property
    def in_channels(self):
        return self._config.in_channels

    @property
    def out_channels(self):
        return self._config.out_channels

    @property
    def embedding_length(self):
        return self._config.embedding_length

    @property
    def d(self):
        return self.embedding_length

    @property
    def skipgram_pattern(self):
        return self._config.skipgram_pattern

    # endregion

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # (B, in_channels, N, d) -> (B, out_channels, N)
        features = self._projected_feature_layer(x).squeeze(dim=-1)

        # (B, in_channels, N, d) -> (B, out_channels, N)
        raw_attention = self._attention_layer(x).squeeze(dim=-1)
        # (B, out_channels, N) -> (B, out_channels, N)
        attention = self._softmax(raw_attention)

        # (B, out_channels, N)
        logits = attention * features

        # (B, out_channels, N) -> (B, out_channels)
        encoding = torch.sum(logits, dim=-1)

        return encoding



import re

DECORATOR_GROUP_PATTERN = r"([^\w\d\s\']+)"
DECORATOR_GROUP_REGEX = re.compile(DECORATOR_GROUP_PATTERN)

DOUBLE_QUOTE_PATTERN = r"\""
DOUBLE_QUOTE_REGEX = re.compile(DOUBLE_QUOTE_PATTERN)

QUOTED_TEXT_PATTERN = r"\'([^\"\'\s]+)\'"
QUOTED_TEXT_REGEX = re.compile(QUOTED_TEXT_PATTERN)

MEANINGLESS_PATTERN = r"^[^\w\d\!\$\%\&\?\.]+$"
MEANINGLESS_REGEX = re.compile(MEANINGLESS_PATTERN)


# Copyright 2021 Miljenko Šuflaj
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from setuptools import setup


def readme():
    with open("README.md") as f:
        return f.read()


setup(
    name="prado",
    version="0.3.1",
    description="The PyTorch implementation of PRADO",
    long_description=readme(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: GPU",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    keywords="prado pytorch torch kaliamoorthi ai",
    url="http://github.com/suflaj/prado",
    author="Miljenko Šuflaj",
    author_email="headsouldev@gmail.com",
    license="Apache License 2.0",
    packages=[
        "prado",
        "prado.datasets",
        "prado.layers",
        "prado.layers.convolutions",
        "prado.layers.embeddings",
        "prado.layers.projections",
    ],
    install_requires=[
        "bitarray",
        "datasketch",
        "masked-convolution",
        "nltk",
        "tqdm",
    ],
    include_package_data=True,
    zip_safe=False,
)



import copy
import sys
from typing import Any, Callable, Dict

from torch.utils.data import Dataset
from tqdm import tqdm

from . import assertion


class ProcessedDataset(Dataset):
    @staticmethod
    def _check_init_args(
        original_dataset: Dataset,
        transformation_map: Dict[int, Callable[[str], Any]],
        verbosity: int,
    ):
        assertion.assert_type("original_dataset", original_dataset, Dataset)
        assertion.assert_type("transformation_map", transformation_map, dict)
        assertion.assert_type("verbosity", verbosity, int)

        assertion.assert_all_int("transfomation_map keys", transformation_map.keys())
        assertion.assert_all_non_negative(
            "transformation_map keys", transformation_map.keys()
        )
        assertion.assert_all_callable(
            "transformation_map values", transformation_map.values()
        )

        return original_dataset, transformation_map, verbosity

    def __init__(
        self,
        original_dataset: Dataset,
        transformation_map: Dict[int, Callable[[str], Any]],
        verbosity: int = 0,
    ):
        super().__init__()

        dataset, transformation_map, verbosity = ProcessedDataset._check_init_args(
            original_dataset=original_dataset,
            transformation_map=transformation_map,
            verbosity=verbosity,
        )

        self._content = list()

        iterator = iter(dataset)

        if verbosity > 0:
            iterator = tqdm(
                iterator,
                desc="Transforming dataset",
                total=len(dataset),
                file=sys.stdout,
            )

        for row in iterator:
            new_content = list(row)

            for index, function in transformation_map.items():
                new_content[index] = function(new_content[index])

            self._content.append(tuple(new_content))

    def __getitem__(self, key):
        return self._content[key]

    def __len__(self):
        return len(self._content)

    def __repr__(self):
        return f"ProcessedDataset(len={len(self)})"

    def __str__(self):
        return repr(self)
